# parameters to be changed for each run
no.interval <- 50
no.maxIteration <- 100
no.initialRate <- 0.3
region.peak <- 0.5
region.border <- 0.5

file.SNP <- "input.SNP"
file.core <- "input.region"
file.model <- ""
file.output <- "output.RHMM"


BaumWelch <- function (model){
	if(no.maxIteration == 0) {
		return(model)
	}
	phylo <- model$phylo
	param <- model$param
	no.branch <- length(phylo$brLen)
	core.length <- phylo$obvLoc[length(phylo$obvLoc)]
	old.LL <- c()
	for ( iter in 1:no.maxIteration) {
		LL <- c()
		evenFreq <- matrix(ncol=2,nrow=no.branch)
		iniMut <- c()
		recAllDensity <- c(0,0)
		recAllEvent <- c(0,0)
		mutAllDensity <- c(0,0)
		recAllLen <- c(0,0)
		all.A <- matrix(0, ncol=2,nrow=2)
		all.B <- matrix(0, ncol=2,nrow=2)
		for ( brId in 1:no.branch) {
			# cat("Running iteration: ", iter, " branch: ",phylo$name[brId], " ( ", brId," / ", no.branch," ) \n")
			branch <- list(obvSeq=phylo$obvSeq[brId,], obvLoc=phylo$obvLoc, name=phylo$name[brId], brLen=phylo$brLen[brId], iniMut=phylo$iniMut[brId])
			if(branch$iniMut < 1e-6) {branch$iniMut <- 1e-6
			} else if(branch$iniMut > 0.999999) {branch$iniMut <- 0.999999}
			if(branch$brLen < 1e-12) {branch$brLen <- 1e-12}
			if(param$recSNP < 1e-12) {param$recSNP <- 1e-12
			} else if (param$recSNP > 0.999999) {param$recSNP <- 0.999999}
			if(param$recLen < 1e-6) {param$recLen <- 1e-6}
			if(param$recFreq < 1e-6) {param$recFreq <-1e-6}
			if(param$mutFreq < 1e-6) {param$mutFreq <-1e-6}

			measures <- EMstep(branch, param)
			
			evenFreq[brId,] <- c(measures$estimateB[1,1]/measures$estimateB[2,1], measures$estimateA[1,2]/sum(measures$estimateA[1,]))
			iniMut[brId] <- measures$iniMut
			
			all.A <- all.A + measures$estimateA
			all.B <- all.B + measures$estimateB
			LL[brId] <- measures$LL
		}
		if ( iter == 1) {diff <- Inf}
		else {diff <- sum(LL - old.LL)}
		cat("Initial rate: ", no.initialRate, "iteration: ", iter, "; LL = ", sum(LL), "; diff = ", diff, "\n")
		cat("old REC: Nu: ", param$recSNP, ", Delta: ", param$recLen,", R: ", param$recFreq*core.length, "; old theta: ",param$mutFreq*core.length,"\n")
		cat("old Branch Length:\n")
		print(phylo$brLen)
#		print(phylo$iniMut)
		
		param$diff <- diff
		if( diff < 0.00001 && iter > 2 ) { break }
		param$recSNP <- all.B[1,2]/all.B[2,2]
		param$recLen <- all.A[2,1]/sum(all.A[2,])
		param$recFreq <- no.branch * all.A[1,2]/sum(all.A[1,])
		param$mutFreq <- no.branch * all.B[1,1]/all.B[2,1]
		phylo$brLen <- 2*(evenFreq[,1] + evenFreq[,2])/(param$recFreq + param$mutFreq)
		
		phylo$iniMut <- iniMut
		old.LL <- LL
		param$LL <- sum(LL)
		cat("new REC: Nu: ", param$recSNP, ", Delta: ", param$recLen,", R: ", param$recFreq*core.length, "; new theta: ",param$mutFreq*core.length,"\n")
		cat("new Branch Length:\n")
		print(phylo$brLen)
	}
	return(list(phylo=phylo, param=param))
}

identifyRegions <- function (model, region.peak, region.border){
	phylo <- model$phylo
	param <- model$param
	no.branch <- length(phylo$brLen)

	LL <- 0
	recombination <- matrix(NA, ncol=5, nrow=0)
	for ( brId in 1:no.branch) {
		cat("Estimating recombination regions in branch: ",phylo$name[brId], " ( ", brId," / ", no.branch," ) \n")
		branch <- list(obvSeq=phylo$obvSeq[brId,], obvLoc=phylo$obvLoc, name=phylo$name[brId], brLen=phylo$brLen[brId], iniMut=phylo$iniMut[brId])
		if(branch$iniMut < 1e-6) {branch$iniMut <- 1e-6
		} else if(branch$iniMut > 0.999999) {branch$iniMut <- 0.999999}
		if(branch$brLen < 1e-12) {branch$brLen <- 1e-12}
		if(param$recSNP < 1e-12) {param$recSNP <- 1e-12
		} else if (param$recSNP > 0.999999) {param$recSNP <- 0.999999}
		if(param$recLen < 1e-6) {param$recLen <- 1e-6}
		if(param$recFreq < 1e-6) {param$recFreq <- 1e-6}
		if(param$mutFreq < 1e-6) {param$mutFreq <- 1e-6}

		regions <- identifyPerRegion(branch, param, region.peak, region.border)
		LL <- LL + regions$LL
		if(length(regions$recomb) == 4) {
			recombination <- rbind(recombination, c(regions$recomb, phylo$name[brId]))
		} else {
			recombination <- rbind(recombination, cbind(regions$recomb, rep(phylo$name[brId],nrow(regions$recomb))))
		}
	}
	cat("LL = ", LL, "\n")
	cat("REC: SNP density: ", param$recSNP, ", Mean length: ", param$recLen,", Frequency: ", param$recFreq*core.length, "; Mutation: ",param$mutFreq*core.length,"\n")
	cat("Branch Length:\n")
	print(phylo$brLen)
	return(list(recombination=recombination, LL=LL))
}

identifyPerRegion <- function(branch, param, region.peak, region.border) {
	no.state <- 2
	n <- length(branch$obvSeq)

	A <- matrix(byrow=TRUE, ncol=2, c(1-param$recFreq*branch$brLen/2, param$recFreq*branch$brLen/2,
									  param$recLen, 1-param$recLen))
	B <- matrix(byrow=TRUE, ncol=2, c(1-param$mutFreq*branch$brLen/2, param$mutFreq*branch$brLen/2, 
									  1-param$recSNP, param$recSNP))
	delta <- c(branch$iniMut, 1-branch$iniMut)
	
	cond <- Estep(branch$obvSeq, branch$obvLoc, A, B, delta)
	u <- cond$u
	Pinterval <- PEstimate(A, B, no.interval)
	peak.sites <- which(u[,2] >= region.peak)
	region <- matrix(0, ncol=5, nrow=0)
	if(length(peak.sites) > 0) {
		for( i in 1:length(peak.sites)) {
			if(nrow(region) > 0 && peak.sites[i] <= region[nrow(region),5]) {
				next
			}
			start.site <- branch$obvLoc[peak.sites[i]]
			end.site <- branch$obvLoc[peak.sites[i]]
			peak.value <- 0
			end.loc <- peak.sites[i]
			for( j in peak.sites[i]:1) {
				if(u[j,2] >= region.border) {
					start.site <- branch$obvLoc[j]
					if(peak.value < u[j,2]) {peak.value <- u[j,2]}
				} else {
					dist <- branch$obvLoc[j+1]-branch$obvLoc[j]
					interval <- u[j,2]*(1-Pinterval[dist,1:dist]) + u[j+1,2]*Pinterval[dist,1:dist]
					start.site <- branch$obvLoc[j] + which(interval >=region.border)[1]
					break
				}
			}
			for( j in (peak.sites[i]+1):n) {
				if(u[j,2] >= region.border) {
					end.site <- branch$obvLoc[j]
					end.loc <- j
					if(peak.value < u[j,2]) {peak.value <- u[j,2]}
				} else {
					dist <- branch$obvLoc[j]-branch$obvLoc[j-1]
					interval <- u[j,2]*(1-Pinterval[dist,1:dist]) + u[j-1,2]*Pinterval[dist,1:dist]
					end.site <- branch$obvLoc[j] - which(interval >=region.border)[1]
					break
				}
			}
			region <- rbind(region, c(start.site, end.site, peak.value, end.site-start.site+1, end.loc))
		}
	}
	return(list(recomb=region[,1:4], LL=cond$LL))
}

EMstep <- function(branch, param) {
	# branch$obvSeq  branch$obvLoc  branch$name  branch$brLen  branch$iniMut
	# param$recSNP  param$recFreq  param$recLen param$LL param$diff
	no.state <- 2
	n <- length(branch$obvSeq)

	A <- matrix(byrow=TRUE, ncol=2, c(1-param$recFreq*branch$brLen/2, param$recFreq*branch$brLen/2,
									  param$recLen, 1-param$recLen))
	B <- matrix(byrow=TRUE, ncol=2, c(1-param$mutFreq*branch$brLen/2, param$mutFreq*branch$brLen/2,
									  1-param$recSNP, param$recSNP))
	delta <- c(branch$iniMut, 1-branch$iniMut)
	
	cond <- Estep(branch$obvSeq, branch$obvLoc, A, B, delta)
	u <- cond$u
	delta <- apply(cond$u, 2, sum)/sum(cond$u)

	Dinterval <- DEstimate(A, B, no.interval)
	tmp <- apply(cbind(u[1:(n-1),1], u[2:n,1], Dinterval[branch$obvLoc[2:n]-branch$obvLoc[1:(n-1)],]), 1, function(x) x[c(order(x[1:2], decreasing = TRUE),3:4)])
	totalsite <- apply(u, 2, sum) + c(sum(tmp[1,]*tmp[3,]+tmp[2,]*tmp[4,]), sum((1-tmp[1,])*tmp[3,]+(1-tmp[2,])*tmp[4,]))
	totalsnp <- apply(branch$obvSeq*u, 2, sum)
	k=which(totalsite < totalsnp)
	if(length(k) > 0) {totalsite[k] <- totalsnp[k]*1.1}
	estimateB <- rbind(totalsnp,totalsite)
	return(list(iniMut=delta[1], estimateB=estimateB, estimateA=cond$estimateA, LL=cond$LL))
}

Estep <- function (obvSeq, obvLoc, A, B, delta) {
	siteLK <- matrix(sapply(obvSeq, function(x) B[,x+1]), ncol=2, byrow=TRUE)
	As <- AEstimate(A, B, no.interval)
	As2 <- As / AEstimate(A, matrix(c(1,1,0,0),ncol=2), no.interval)
    m <- ncol(siteLK)
    n <- nrow(siteLK)
    y <- forwardback(siteLK, obvLoc, As, delta)
	logprob <- log(siteLK)
    LL <- y$LL
	interval <- obvLoc[2:n]-obvLoc[1:(n-1)]
    v <- array(NA, dim = c(n - 1, m, m))
    for (k in 1:m) {
		logPi <- log(t(sapply(interval, function(i) As2[i,,k]*A[,k]*i)))
        logPbeta <- matrix(logprob[-1, k] + y$logbeta[-1, k], byrow = FALSE, nrow = n - 1, ncol = m)
        v[, , k] <- logPi + y$logalpha[-n, ] + logPbeta - LL
    }
	v <- exp(v)
	
	Pi <- apply(v, MARGIN = c(2, 3), FUN = sum)
	rm(v)
	gc()
	u <- exp(y$logalpha + y$logbeta - LL)
    return(list(u = u, estimateA = Pi, LL = LL))
}

AEstimate <- function(A, B, no.interval) {
	no.state <- ncol(A)
	transition <- array(NA, dim=c(no.interval, no.state, no.state))
	transition[1,,] <- A
	if(no.interval > 1) {
		for( i in 2:no.interval) {
			for( j in 1:no.state) {
				transition[i,j,] <- (transition[i-1,j,] %*% A) * B[,1]
			}
		}
	}
	transition[transition<=1e-50] <- 1e-50
	return (transition)
}

DEstimate <- function(A, B, no.interval) {
	no.state <- ncol(A)
	transitSite <- matrix(0, nrow=no.interval, ncol=no.state)
	for ( itv in 1:no.interval) {
		psum <- 0
		wsum <- 0
		for ( i in 1:itv) {
			p <- (A[1,1]*B[1,1])^(i-1)*(A[2,2]*B[2,1])^(itv-i)
			psum <- psum + p
			wsum <- wsum + (i-1)*p
		}
		transitSite[itv,] <- c(wsum/psum, i-1-wsum/psum)
	}
	return (transitSite)
}

PEstimate <- function(A, B, no.interval) {
	no.state <- ncol(A)
	transitSite <- matrix(0, nrow=no.interval, ncol=no.interval)
	for ( itv in 1:no.interval) {
		psum <- c()
		for ( i in 1:itv) {
			p <- (A[1,1]*B[1,1])^(i-1)*(A[2,2]*B[2,1])^(itv-i)
			if(i == 1) {psum[i] <- p}
			else {psum[i] <- psum[i-1] + p}
		}
		transitSite[itv,1:length(psum)] <- psum/psum[length(psum)]
	}
	return (transitSite)
}

forwardback <- function(prob, site, As, delta){
    m <- ncol(prob)
    n <- nrow(prob)
    #   forward probabilities alpha_ij
    phi <- as.double(delta)
    logalpha <- matrix(as.double(rep(0, m*n)), nrow=n)
    lscale <- as.double(0)
    
	#  loop1 using R code
	for (i in 1:n){
		if (i > 1) phi <- phi %*% As[site[i]-site[i-1],,]
		phi <- phi*prob[i,]
		sumphi <- sum(phi)
		phi <- phi/sumphi
		lscale <- lscale + log(sumphi)
		logalpha[i,] <- log(phi) + lscale
	}
	LL <- lscale
    
    #   backward probabilities beta_ij
    logbeta <- matrix(as.double(rep(0, m*n)), nrow=n)
    phi <- as.double(rep(1/m, m))
    lscale <- as.double(log(m))
    
	#  loop2 using R code
	for (i in seq(n-1, 1, -1)){
		phi <- As[site[i+1]-site[i],,] %*% (prob[i+1,]*phi)
		logbeta[i,] <- log(phi) + lscale
		sumphi <- sum(phi)
		phi <- phi/sumphi
		lscale <- lscale + log(sumphi)
	}
    
    return(list(logalpha=logalpha, logbeta=logbeta, LL=LL))
}

readSNPs <- function (file.SNP, file.core) {
	core.genome <- matrix(scan(file.core, what = double(0), n = 99999999), byrow=TRUE, ncol=2)
	core.length <- sum(apply(core.genome, 1, function(x) x[2]-x[1]+1))
	branch.snp <- matrix(NA, ncol=3, nrow=0)
	fin <- file(file.SNP, "r")
	while (length(oneLine <- readLines(fin, n = 1, warn = FALSE)) > 0) {
		snp <- strsplit(oneLine, " ")[[1]]
		for ( i in 2:length(snp)) {
			branch.snp <- rbind(branch.snp, c(snp[i], snp[1], snp[1]))
		}
	}
	close(fin)
	
	branch.snp <- branch.snp[order(as.numeric(branch.snp[,3])),]
	core.site <- 0
	j <- 1
	for ( i in 1:nrow(branch.snp)) {
		while (as.numeric(branch.snp[i,3]) > core.genome[j,2]) {
			core.site <- core.site + core.genome[j,2] - core.genome[j,1] + 1
			j <- j + 1
		}
		
		if (as.numeric(branch.snp[i,3]) < core.genome[j,1]) {
			warning("Site: ", branch.snp[i,3], " in branch: ", branch.snp[i,1], " is not in the core genome.")
			branch.snp[i,2] <- -1
		} else {
			branch.snp[i,2] <- as.numeric(branch.snp[i,3]) - core.genome[j,1]+1 + core.site
		}
	}
	obvLoc <- sort(unique(c(seq(1, core.length, no.interval), core.length, as.numeric(branch.snp[branch.snp[,2] != "-1",2]))))
#	obvLoc <- obvLoc[obvLoc >= 1]
#	obvLoc <- obvLoc[obvLoc <= core.length]
	brName <- sort(unique(branch.snp[,1]))
	obvSeq <- matrix(0, nrow=length(brName), ncol=length(obvLoc), dimnames=list(branch=brName, refSite=obvLoc))
	for ( name in brName) {
		brSNP <- branch.snp[branch.snp[,1] == name,2]
		obvSeq[name, brSNP[brSNP != "-1"]] <- 1
	}
	weight <- c()
	for ( i in 1:length(brName)) {
		no.snp <- sum(obvSeq[i,])
		weight[i] <- (-log(no.snp/core.length))*no.snp + (-log((core.length-no.snp)/core.length))*(core.length-no.snp)
	}
	return(list(name=brName, weight=weight, obvLoc=obvLoc, obvSeq=obvSeq, brLen=rep(1, length(name)), iniMut=rep(1,length(name)), oriSNP=branch.snp, coreGenome=core.genome))
}

initialModel <- function (phylo, iniLevel) {
	#iniLevel <- c(0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9)
	no.branch <- length(phylo$name)
	core.length <- phylo$obvLoc[length(phylo$obvLoc)]
	spacer <- matrix(0, ncol=3, nrow=0)
	for( i in 1:no.branch) {
		muts <- unique(c(1, as.numeric(phylo$oriSNP[phylo$oriSNP[,1] == phylo$name[i],2])))
		muts <- sort(muts[muts > 0])
		muts[1] <- 1 - (core.length-muts[length(muts)])
		for( j in 2:length(muts)) {
			spacer <- rbind(spacer, c(i, j-1, muts[j]-muts[j-1]))
		}
	}
	spacer <- spacer[order(spacer[, 3]), ]
	criteria <- spacer[ceiling(nrow(spacer)*iniLevel), 3]

	spacer <- spacer[order(spacer[,2]), ]
	spacer <- spacer[order(spacer[,1]), ]
	brLen <- rep(0, no.branch)
	iniMut <- rep(0, no.branch)
	recSNP <- 0
	recLen <- 0
	recFreq <- 0
	mutFreq <- 0
	
	recomb <- matrix(0, nrow=0, ncol=3)
	for(j in 1:nrow(spacer)) {
		if(spacer[j,3] > criteria) {
			brLen[spacer[j,1]] <- brLen[spacer[j,1]] + 1
		} else {
			if(j > 1 && spacer[j-1,1] == spacer[j,1] && spacer[j-1,3] <= criteria) {
				recomb[nrow(recomb),2] <- recomb[nrow(recomb),2] + spacer[j,3]
				recomb[nrow(recomb),3] <- recomb[nrow(recomb),3] + 1
			} else {
				recomb <- rbind(recomb, c(spacer[j,1], spacer[j,3], 1))
				# brLen[spacer[j,1]] <- brLen[spacer[j,1]] - 1
			}
		}
		iniMut[spacer[j,1]] <- 0.5
	}
	recSNP <- sum(recomb[,3])/sum(recomb[,2])
	recLen <- 1/mean(recomb[,2])
	recFreq <- nrow(recomb)/core.length
	mutFreq <- (sum(brLen) - nrow(recomb))/core.length
	brLen <- 2*brLen/(mutFreq+recFreq)/core.length
	phylo$brLen <- brLen
	phylo$iniMut <- iniMut
	
	return(list(phylo=phylo, param=list(mutFreq=mutFreq, recSNP=recSNP, recLen=recLen, recFreq=recFreq, LL=-Inf, diff=Inf)))
}

readModel <- function(phylo, file.model) {
	fin <- file(file.model, "r")
	while (length(oneLine <- readLines(fin, n = 1, warn = FALSE)) > 0) {
		if(oneLine == "theta:") {
			oneLine <- readLines(fin, n = 1, warn = FALSE)
			mutFreq <- as.numeric(oneLine)
		} else if(oneLine == "R:") {
			oneLine <- readLines(fin, n = 1, warn = FALSE)
			recFreq <- as.numeric(oneLine)
		} else if (oneLine == "delta:") {
			oneLine <- readLines(fin, n = 1, warn = FALSE)
			recLen <- as.numeric(oneLine)
		} else if (oneLine == "nu:") {
			oneLine <- readLines(fin, n = 1, warn = FALSE)
			recSNP <- as.numeric(oneLine)
		} else if (oneLine == "Branch Length:") {
			oneLine <- readLines(fin, n = 1, warn = FALSE)
			phylo$brLen <- as.numeric(strsplit(oneLine, " ")[[1]])
		} else if (oneLine == "u':") {
			oneLine <- readLines(fin, n = 1, warn = FALSE)
			phylo$iniMut <- as.numeric(strsplit(oneLine, " ")[[1]])
		}
	}
	close(fin)
	
	return(list(phylo=phylo, param=list(mutFreq=mutFreq, recSNP=recSNP, recLen=recLen, recFreq=recFreq, LL=-Inf, diff=Inf)))
}

core2ref <- function(coreGenome, rec) {
	ref.table <- matrix(0,nrow=nrow(coreGenome),ncol=4)
	ref.table[,1:2] <- coreGenome
	for( i in 1:nrow(ref.table)) {
		if(i == 1) {ref.table[i,3:4] <- c(1,ref.table[i,2]-ref.table[i,1]+1)}
		else {ref.table[i,3:4] <- c(ref.table[i-1,4]+1, ref.table[i,2]-ref.table[i,1]+1 + ref.table[i-1,4])}
	}
	sites <- unique(as.array(as.numeric(rec$recombination[,1:2])))
	names(sites) <- sites
	j <- 1
	for(i in order(sites)) {
		site <- sites[i]
		while(site > ref.table[j,4]) {
			j <- j + 1
		}
		sites[i] <- site - ref.table[j,3] + ref.table[j,1]
	}
	rec$recombination[,1:2] <- as.character(sites[as.character(rec$recombination[,1:2])])
	return(rec)
}

options(echo=FALSE) # if you want see commands in output file
args <- commandArgs(trailingOnly = TRUE)
if(length(args) < 3) {
	cat("Rscript RHMM.R <file.SNP> <file.coreGenome> <file.output> <initial R/M rate (0.3) or read model> <Border of recomb region (0.5)> <Peak of recomb region (0.5)> <Maximum no. of iterations (100)> <no. of sites between reference sites (50)>\n")
	quit()
} else {
	if(length(args) >= 1) {file.SNP <- args[1]}
	if(length(args) >= 2) {file.core <- args[2]}
	if(length(args) >= 3) {file.output <- args[3]}
	if(length(args) >= 4) {
		if(length(grep("^[0-9\\.]+$", args[4], perl=TRUE))>0) {
			no.initialRate <- as.numeric(args[4])
		} else {
			file.model <- args[4]
			no.initialRate <- file.model
		}
	}
	if(length(args) >= 5) {region.border <- as.numeric(args[5])}
	if(length(args) >= 6) {region.peak <- as.numeric(args[6])}
	if(length(args) >= 7) {no.maxIteration <- as.numeric(args[7])}
	if(length(args) >= 8) {no.interval <- round(as.numeric(args[8]))}
}

align <- readSNPs(file.SNP, file.core)
core.length <- align$obvLoc[length(align$obvLoc)]
if(file.model == "") {
	model <- initialModel(align, no.initialRate)
} else {
	model <- readModel(align, file.model)
}
model <- BaumWelch(model)

options(digits=15)
fout <- file(file.output, "w")
open(fout)
write("Model initial:", file=fout)
write(no.initialRate, file=fout)
write("Model LL, Diff:", file=fout)
write(c(model$param$LL, model$param$diff), file=fout, ncol=2)
write("theta:", file=fout)
write(model$param$mutFreq * core.length, file=fout)
write("R:", file=fout)
write(model$param$recFreq * core.length, file=fout)
write("delta:", file=fout)
write(model$param$recLen, file=fout)
write("nu:", file=fout)
write(model$param$recSNP, file=fout)
write("Branch Names:", file=fout)
write(model$phylo$name, file=fout, ncol=length(model$phylo$name))
write("Branch Length:", file=fout)
write(model$phylo$brLen, file=fout, ncol=length(model$phylo$brLen))
write("u':", file=fout)
write(model$phylo$iniMut, file=fout, ncol=length(model$phylo$iniMut))

rec <- identifyRegions(model, region.peak, region.border)
rec <- core2ref(model$phylo$coreGenome, rec)
write("Recombinant stretches:", file=fout) 
write(t(rec$recombination), file=fout, ncol=ncol(rec$recombination))
close(fout)
